package io.stratosfy.stratosfy_esp_ble_provisioning


import android.Manifest
import android.bluetooth.BluetoothDevice
import android.bluetooth.le.ScanResult
import android.content.Context
import android.content.pm.PackageManager
import android.os.Build
import android.util.Log
import androidx.annotation.NonNull
import androidx.annotation.RequiresApi
import androidx.core.app.ActivityCompat
import com.espressif.provisioning.ESPConstants
import com.espressif.provisioning.ESPProvisionManager
import com.espressif.provisioning.device_scanner.BleScanner
import com.espressif.provisioning.listeners.BleScanListener
import com.espressif.provisioning.listeners.ProvisionListener
import io.flutter.embedding.engine.plugins.FlutterPlugin
import io.flutter.embedding.engine.plugins.activity.ActivityAware
import io.flutter.embedding.engine.plugins.activity.ActivityPluginBinding
import io.flutter.plugin.common.MethodCall
import io.flutter.plugin.common.MethodChannel
import io.flutter.plugin.common.MethodChannel.MethodCallHandler
import io.flutter.plugin.common.MethodChannel.Result
import java.time.LocalDateTime


/** EspBluetoothProvisioningPlugin */
class EspBluetoothProvisioningPlugin : FlutterPlugin, MethodCallHandler, ActivityAware {
    private lateinit var channel: MethodChannel
    private lateinit var context: Context
    private var bleScanner: BleScanner? = null
    private val deviceNamePrefix: String? = null
    private lateinit var permissionManager: PermissionManager
    private lateinit var provisionManager: ESPProvisionManager
    private var devices = ArrayList<BleDevice>()
    private var devicesList = ArrayList<BleTooth>()
    private var provisionStatus: String? = null
    private var ssid: String? = null
    private var password: String? = null
    private var pin: String? = null
    private var selectedDevice: BleDevice? = null
    private var selectedDeviceName: String? = null
    private var isAndroid12: Boolean = Build.VERSION.SDK_INT >= Build.VERSION_CODES.S

    override fun onAttachedToEngine(@NonNull flutterPluginBinding: FlutterPlugin.FlutterPluginBinding) {
        channel = MethodChannel(flutterPluginBinding.binaryMessenger, "esp_ble_provisioning")
        channel.setMethodCallHandler(this)
    }

    private fun initializeWithContext() {
        permissionManager = PermissionManager(context)
        provisionManager = ESPProvisionManager.getInstance(context)
    }

    private fun checkPermissions(): Boolean {
        if (isAndroid12) {
            return (ActivityCompat.checkSelfPermission(
                context,
                Manifest.permission.ACCESS_FINE_LOCATION
            ) == PackageManager.PERMISSION_GRANTED
                    && ActivityCompat.checkSelfPermission(
                context,
                Manifest.permission.BLUETOOTH
            ) == PackageManager.PERMISSION_GRANTED
                    && ActivityCompat.checkSelfPermission(
                context,
                Manifest.permission.BLUETOOTH_ADMIN
            ) == PackageManager.PERMISSION_GRANTED
                    && ActivityCompat.checkSelfPermission(
                context,
                Manifest.permission.BLUETOOTH_SCAN
            ) == PackageManager.PERMISSION_GRANTED)
        } else {
            return (ActivityCompat.checkSelfPermission(
                context,
                Manifest.permission.ACCESS_FINE_LOCATION
            ) == PackageManager.PERMISSION_GRANTED
                    && ActivityCompat.checkSelfPermission(
                context,
                Manifest.permission.BLUETOOTH
            ) == PackageManager.PERMISSION_GRANTED
                    && ActivityCompat.checkSelfPermission(
                context,
                Manifest.permission.BLUETOOTH_ADMIN
            ) == PackageManager.PERMISSION_GRANTED)
        }
    }

    //    @RequiresApi(Build.VERSION_CODES.O)
    @RequiresApi(Build.VERSION_CODES.S)
    override fun onMethodCall(@NonNull call: MethodCall, @NonNull result: Result) {
        println(Build.VERSION.SDK_INT)
        println(Build.VERSION_CODES.S)
        println(isAndroid12)

        val method: String = call.method
        when {
            method == ("getPlatformVersion") -> {
                result.success("Android " + Build.VERSION.RELEASE)
            }

            method == ("scanBleDevice") -> {
                Log.e("BLEProvisionLanding", "Scan Device")
                if (ActivityCompat.checkSelfPermission(
                        context,
                        Manifest.permission.ACCESS_FINE_LOCATION
                    ) == PackageManager.PERMISSION_GRANTED
                    && ActivityCompat.checkSelfPermission(
                        context,
                        Manifest.permission.BLUETOOTH
                    ) == PackageManager.PERMISSION_GRANTED
                    && ActivityCompat.checkSelfPermission(
                        context,
                        Manifest.permission.BLUETOOTH_ADMIN
                    ) == PackageManager.PERMISSION_GRANTED
                ) {
                    if (isAndroid12) {
                        if (ActivityCompat.checkSelfPermission(
                                context,
                                Manifest.permission.BLUETOOTH_SCAN
                            ) == PackageManager.PERMISSION_GRANTED
                        ) {
                            Log.e("BLEProvisionLanding", "Searching for devices")
                            provisionManager = ESPProvisionManager.getInstance(context)
                            devices.clear()
                            devicesList.clear()
                            provisionManager.searchBleEspDevices(
                                deviceNamePrefix,
                                object : BleScanListener {

                                    override fun scanStartFailed() {
                                        Log.d(
                                            "BLEProvisionLanding",
                                            "Please turn on Bluetooth to connect BLE device"
                                        )
                                    }

                                    override fun onPeripheralFound(
                                        device: BluetoothDevice,
                                        scanResult: ScanResult
                                    ) {
                                        Log.d("BLEProvisionLanding", "BLE device $device")

                                        val dev = BleDevice()
                                        val ble = BleTooth()
                                        dev.name = scanResult.scanRecord!!.deviceName.toString()
                                        ble.name = scanResult.scanRecord!!.deviceName.toString()

                                        if (scanResult.scanRecord!!.serviceUuids != null && scanResult.scanRecord!!.serviceUuids.size > 0) {
                                            dev.uuid =
                                                scanResult.scanRecord!!.serviceUuids[0].toString()
                                            ble.uuid =
                                                scanResult.scanRecord!!.serviceUuids[0].toString()
                                        }
                                        dev.bluetoothDevice = (device)
                                        Log.d("BLEProvisionLanding", "OPF--> dev: $dev");
                                        Log.d("BLEProvisionLanding", "OPF--> ble: $ble");

                                        if (!devices.contains(dev)) {
                                            devices.add(dev)
                                        }
                                        if (!devicesList.contains(ble)) {
                                            devicesList.add(ble)
                                        }
                                        if (selectedDeviceName == dev.uuid) {
                                            selectedDevice = dev
                                            Log.d(
                                                "BLEProvisionLanding",
                                                "Selected BLE device $selectedDevice"
                                            )
                                        }
//                          sendDeviceResponse(method, devicesList)
                                    }

                                    override fun scanCompleted() {
                                        sendDeviceResponse(method, devicesList)
                                        Log.d("BLEProvisionLanding", "Scan completed")
                                    }

                                    override fun onFailure(e: Exception) {
                                        Log.e("BLEProvisionLanding", e.message!!)
                                        e.printStackTrace()
                                    }
                                })
                        } else {
                            if (!permissionManager.hasLocationPermission()) {
                                permissionManager.requestLocationPermission()
                            }
                        }
                    } else {
                        Log.e("BLEProvisionLanding", "Searching for devices")
                        provisionManager = ESPProvisionManager.getInstance(context)
                        devices.clear()
                        devicesList.clear()
                        provisionManager.searchBleEspDevices(
                            deviceNamePrefix,
                            object : BleScanListener {

                                override fun scanStartFailed() {
                                    Log.d(
                                        "BLEProvisionLanding",
                                        "Please turn on Bluetooth to connect BLE device"
                                    )
                                }

                                override fun onPeripheralFound(
                                    device: BluetoothDevice,
                                    scanResult: ScanResult
                                ) {
                                    Log.d("BLEProvisionLanding", "BLE device $device")

                                    val dev = BleDevice()
                                    val ble = BleTooth()
                                    dev.name = scanResult.scanRecord!!.deviceName.toString()
                                    ble.name = scanResult.scanRecord!!.deviceName.toString()

                                    if (scanResult.scanRecord!!.serviceUuids != null && scanResult.scanRecord!!.serviceUuids.size > 0) {
                                        dev.uuid =
                                            scanResult.scanRecord!!.serviceUuids[0].toString()
                                        ble.uuid =
                                            scanResult.scanRecord!!.serviceUuids[0].toString()
                                    }
                                    dev.bluetoothDevice = (device)
                                    Log.d("BLEProvisionLanding", "OPF--> dev: $dev");
                                    Log.d("BLEProvisionLanding", "OPF--> ble: $ble");
//                            try {
                                    if (!devices.contains(dev)) {
                                        devices.add(dev)
                                    }
                                    if (!devicesList.contains(ble)) {
                                        devicesList.add(ble)
                                    }
                                    if (selectedDeviceName == dev.uuid) {
                                        selectedDevice = dev
                                        Log.d(
                                            "BLEProvisionLanding",
                                            "Selected BLE device $selectedDevice"
                                        )
                                    }
//                            } catch (e: Exception) {
//                                Log.d("BLEProvisionLanding",
//                                    "error -->  $e")
//                            }
//                          sendDeviceResponse(method, devicesList)
                                }

                                override fun scanCompleted() {
                                    sendDeviceResponse(method, devicesList)
                                    Log.d("BLEProvisionLanding", "Scan completed")
                                }

                                override fun onFailure(e: Exception) {
                                    Log.e("BLEProvisionLanding", e.message!!)
                                    e.printStackTrace()
                                }
                            })

                    }
                } else {
                    Log.e(
                        "BLEProvisionLanding",
                        "Not able to start scan as Location permission is not granted. Failed!"
                    )
                    if (!permissionManager.hasLocationPermission())
                        permissionManager.requestLocationPermission()
                }
                result.success(true)

            }

            method == ("connectBleDevice") -> {

                selectedDeviceName = call.argument("name")!!
                pin = call.argument("pin")!!

                Log.e("BLEProvisionLanding", "Connect to device Id $selectedDeviceName")

                devices.clear()
                devicesList.clear()
                if (ActivityCompat.checkSelfPermission(
                        context,
                        Manifest.permission.ACCESS_FINE_LOCATION
                    ) == PackageManager.PERMISSION_GRANTED
                    && ActivityCompat.checkSelfPermission(
                        context,
                        Manifest.permission.BLUETOOTH
                    ) == PackageManager.PERMISSION_GRANTED
                    && ActivityCompat.checkSelfPermission(
                        context,
                        Manifest.permission.BLUETOOTH_ADMIN
                    ) == PackageManager.PERMISSION_GRANTED
                ) {
                    if (isAndroid12) {
                        if (
                            ActivityCompat.checkSelfPermission(
                                context,
                                Manifest.permission.BLUETOOTH_SCAN
                            ) == PackageManager.PERMISSION_GRANTED
                            && ActivityCompat.checkSelfPermission(
                                context,
                                Manifest.permission.BLUETOOTH_CONNECT
                            ) == PackageManager.PERMISSION_GRANTED
                        ) {
                            Log.e("BLEProvisionLanding", "Searching for devices")
                            provisionManager.searchBleEspDevices(
                                deviceNamePrefix,
                                object : BleScanListener {

                                    override fun scanStartFailed() {
                                        Log.d(
                                            "BLEProvisionLanding",
                                            "Please turn on Bluetooth to connect BLE device"
                                        )
                                    }

                                    override fun onPeripheralFound(
                                        device: BluetoothDevice,
                                        scanResult: ScanResult
                                    ) {
                                        Log.d("BLEProvisionLanding", "BLE device $device")
                                        Log.e(
                                            "BLEProvisionLanding",
                                            "Uuids" + scanResult.scanRecord!!.serviceUuids.toString()
                                        )
                                        Log.e(
                                            "BLEProvisionLanding",
                                            "Advertising flag   " + scanResult.scanRecord!!.advertiseFlags
                                        )
                                        Log.e(
                                            "BLEProvisionLanding",
                                            "Name   " + scanResult.scanRecord!!.deviceName
                                        )
                                        Log.e(
                                            "BLEProvisionLanding",
                                            "Service data" + scanResult.scanRecord!!.serviceData
                                        )
                                        Log.e(
                                            "BLEProvisionLanding",
                                            "Manufacturer data" + scanResult.scanRecord!!.manufacturerSpecificData
                                        )
                                        Log.e(
                                            "BLEProvisionLanding",
                                            "Advertising flag" + scanResult.scanRecord!!
                                        )

                                        val dev = BleDevice()
                                        val ble = BleTooth()
                                        dev.name = scanResult.scanRecord!!.deviceName.toString()
                                        ble.name = scanResult.scanRecord!!.deviceName.toString()


                                        if (scanResult.scanRecord!!.serviceUuids != null && scanResult.scanRecord!!.serviceUuids.size > 0) {
                                            dev.uuid =
                                                scanResult.scanRecord!!.serviceUuids[0].toString()
                                            ble.uuid =
                                                scanResult.scanRecord!!.serviceUuids[0].toString()
                                        }
                                        dev.bluetoothDevice = (device)

                                        if (!devices.contains(dev)) {
                                            devices.add(dev)
                                        }
                                        if (!devicesList.contains(ble)) {
                                            devicesList.add(ble)
                                        }
                                        if (selectedDeviceName == dev.name) {
                                            selectedDevice = dev
                                            Log.d(
                                                "BLEProvisionLanding",
                                                "Selected BLE device $selectedDevice"
                                            )
                                        }
                                    }

                                    override fun scanCompleted() {
                                        Log.d("BLEProvisionLanding", "Scan completed")
                                        if (selectedDeviceName != null) {
                                            provisionManager.createESPDevice(
                                                ESPConstants.TransportType.TRANSPORT_BLE,
                                                ESPConstants.SecurityType.SECURITY_1
                                            )


                                            Log.d(
                                                "BLEProvisionLanding",
                                                "=================== Connect to device : ${selectedDevice?.bluetoothDevice}"
                                            )
                                            Log.d(
                                                "BLEProvisionLanding",
                                                "=================== Connect to device : " + selectedDevice?.name + " UUID : " + selectedDevice?.uuid
                                            )

                                            if (ActivityCompat.checkSelfPermission(
                                                    context,
                                                    Manifest.permission.ACCESS_FINE_LOCATION
                                                ) == PackageManager.PERMISSION_GRANTED
                                            ) {
                                                if (isAndroid12) {
                                                    if (ActivityCompat.checkSelfPermission(
                                                            context,
                                                            Manifest.permission.BLUETOOTH_SCAN
                                                        ) == PackageManager.PERMISSION_GRANTED
                                                        && ActivityCompat.checkSelfPermission(
                                                            context,
                                                            Manifest.permission.BLUETOOTH_CONNECT
                                                        ) == PackageManager.PERMISSION_GRANTED
                                                    ) {
                                                        Log.e(
                                                            "BLEProvisionLanding",
                                                            "Before Connecting ------------------>."
                                                        )
                                                        try {
                                                            provisionManager.espDevice?.connectBLEDevice(
                                                                selectedDevice?.bluetoothDevice,
                                                                selectedDevice?.uuid
                                                            )
                                                            channel.invokeMethod(method, "success")
                                                            Log.d(
                                                                "BLEProvisionLanding",
                                                                "=================== Connected to device"
                                                            )
                                                        } catch (e: Exception) {
                                                            channel.invokeMethod(method, "failure")
                                                        }
                                                        Log.e(
                                                            "BLEProvisionLanding",
                                                            "After Connecting ------------------>."
                                                        )
                                                    } else {
                                                        result.error(
                                                            "400",
                                                            "Failed to connect to device",
                                                            "Error connecting to device"
                                                        )
                                                        channel.invokeMethod(method, "failure")
                                                    }
                                                } else {
                                                    Log.e(
                                                        "BLEProvisionLanding",
                                                        "Before Connecting ------------------>."
                                                    )
                                                    try {
                                                        provisionManager.espDevice?.connectBLEDevice(
                                                            selectedDevice?.bluetoothDevice,
                                                            selectedDevice?.uuid
                                                        )
                                                        channel.invokeMethod(method, "success")
                                                        Log.d(
                                                            "BLEProvisionLanding",
                                                            "=================== Connected to device"
                                                        )
                                                    } catch (e: Exception) {
                                                        channel.invokeMethod(method, "failure")
                                                    }
                                                    Log.e(
                                                        "BLEProvisionLanding",
                                                        "After Connecting ------------------>."
                                                    )


                                                }
                                            } else {
                                                result.error(
                                                    "400",
                                                    "Failed to connect to device",
                                                    "Error connecting to device"
                                                )
                                                channel.invokeMethod(method, "failure")
                                            }
                                            Log.e(
                                                "BLEProvisionLanding",
                                                "Before Security check ------------------>. $pin"
                                            )
                                            try {
                                                provisionManager.espDevice.proofOfPossession = pin
                                                channel.invokeMethod(method, "success")
                                            } catch (e: Exception) {
                                                channel.invokeMethod(method, "failure")
                                            }
                                            Log.e(
                                                "BLEProvisionLanding",
                                                "After Security check ------------------>."
                                            )

                                        } else {
                                            channel.invokeMethod(method, "failure")
                                        }
                                    }

                                    override fun onFailure(e: Exception) {
                                        Log.e("BLEProvisionLanding", e.message!!)
                                        e.printStackTrace()
                                    }
                                })
                        } else {
                            Log.e(
                                "BLEProvisionLanding",
                                "Not able to start scan as Location permission is not granted. Failed!"
                            )
                        }
                    } else {
                        Log.e("BLEProvisionLanding", "Searching for devices")
                        provisionManager.searchBleEspDevices(
                            deviceNamePrefix,
                            object : BleScanListener {

                                override fun scanStartFailed() {
                                    Log.d(
                                        "BLEProvisionLanding",
                                        "Please turn on Bluetooth to connect BLE device"
                                    )
                                }

                                override fun onPeripheralFound(
                                    device: BluetoothDevice,
                                    scanResult: ScanResult
                                ) {
                                    Log.d("BLEProvisionLanding", "BLE device $device")
                                    Log.e(
                                        "BLEProvisionLanding",
                                        "Uuids" + scanResult.scanRecord!!.serviceUuids.toString()
                                    )
                                    Log.e(
                                        "BLEProvisionLanding",
                                        "Advertising flag   " + scanResult.scanRecord!!.advertiseFlags
                                    )
                                    Log.e(
                                        "BLEProvisionLanding",
                                        "Name   " + scanResult.scanRecord!!.deviceName
                                    )
                                    Log.e(
                                        "BLEProvisionLanding",
                                        "Service data" + scanResult.scanRecord!!.serviceData
                                    )
                                    Log.e(
                                        "BLEProvisionLanding",
                                        "Manufacturer data" + scanResult.scanRecord!!.manufacturerSpecificData
                                    )
                                    Log.e(
                                        "BLEProvisionLanding",
                                        "Advertising flag" + scanResult.scanRecord!!
                                    )

                                    val dev = BleDevice()
                                    val ble = BleTooth()
                                    dev.name = scanResult.scanRecord!!.deviceName.toString()
                                    ble.name = scanResult.scanRecord!!.deviceName.toString()


                                    if (scanResult.scanRecord!!.serviceUuids != null && scanResult.scanRecord!!.serviceUuids.size > 0) {
                                        dev.uuid =
                                            scanResult.scanRecord!!.serviceUuids[0].toString()
                                        ble.uuid =
                                            scanResult.scanRecord!!.serviceUuids[0].toString()
                                    }
                                    dev.bluetoothDevice = (device)

                                    if (!devices.contains(dev)) {
                                        devices.add(dev)
                                    }
                                    if (!devicesList.contains(ble)) {
                                        devicesList.add(ble)
                                    }
                                    if (selectedDeviceName == dev.name) {
                                        selectedDevice = dev
                                        Log.d(
                                            "BLEProvisionLanding",
                                            "Selected BLE device $selectedDevice"
                                        )
                                    }
                                }

                                override fun scanCompleted() {
                                    Log.d("BLEProvisionLanding", "Scan completed")
                                    if (selectedDeviceName != null) {
                                        provisionManager.createESPDevice(
                                            ESPConstants.TransportType.TRANSPORT_BLE,
                                            ESPConstants.SecurityType.SECURITY_1
                                        )


                                        Log.d(
                                            "BLEProvisionLanding",
                                            "=================== Connect to device : ${selectedDevice?.bluetoothDevice}"
                                        )
                                        Log.d(
                                            "BLEProvisionLanding",
                                            "=================== Connect to device : " + selectedDevice?.name + " UUID : " + selectedDevice?.uuid
                                        )

                                        if (ActivityCompat.checkSelfPermission(
                                                context,
                                                Manifest.permission.ACCESS_FINE_LOCATION
                                            ) == PackageManager.PERMISSION_GRANTED
                                        ) {
                                            Log.e(
                                                "BLEProvisionLanding",
                                                "Before Connecting ------------------>."
                                            )
                                            try {
                                                provisionManager.espDevice?.connectBLEDevice(
                                                    selectedDevice?.bluetoothDevice,
                                                    selectedDevice?.uuid
                                                )
                                                channel.invokeMethod(method, "success")
                                                Log.d(
                                                    "BLEProvisionLanding",
                                                    "=================== Connected to device"
                                                )
                                            } catch (e: Exception) {
                                                channel.invokeMethod(method, "failure")
                                            }
                                            Log.e(
                                                "BLEProvisionLanding",
                                                "After Connecting ------------------>."
                                            )


                                        } else {
                                            result.error(
                                                "400",
                                                "Failed to connect to device",
                                                "Error connecting to device"
                                            )
                                            channel.invokeMethod(method, "failure")

                                        }


                                        Log.e(
                                            "BLEProvisionLanding",
                                            "Before Security check ------------------>. $pin"
                                        )
                                        try {
                                            provisionManager.espDevice.proofOfPossession = pin
                                            channel.invokeMethod(method, "success")
                                        } catch (e: Exception) {
                                            channel.invokeMethod(method, "failure")
                                        }
                                        Log.e(
                                            "BLEProvisionLanding",
                                            "After Security check ------------------>."
                                        )

                                    } else {
                                        channel.invokeMethod(method, "failure")
                                    }
                                }

                                override fun onFailure(e: Exception) {
                                    Log.e("BLEProvisionLanding", e.message!!)
                                    e.printStackTrace()
                                }
                            })
                    }
                } else {
                    Log.e(
                        "BLEProvisionLanding",
                        "Not able to start scan as Location permission is not granted. Failed!"
                    )
                }
                result.success(true)
            }

            method == "provisionDevice" -> {
                ssid = call.argument("ssid")!!
                password = call.argument("password")!!

                println("cred $ssid and $password")
                Log.e("BLEProvisioningLanding", "Provisioning cred $ssid and $password")
                Log.e("BLEProvisioningLanding", "Provisioning start" + LocalDateTime.now())
                println("Provisioning start" + LocalDateTime.now())


                //**********PROVISIONING PART

                provisionManager.espDevice.provision(ssid, password, object : ProvisionListener {
                    override fun createSessionFailed(e: java.lang.Exception) {
                        provisionStatus = "CREATE_SESSION_FAILED"
//                      sendProvisionResponse(method, provisionStatus!!)
                        println("createSessionFailed!!" + LocalDateTime.now())
                        Log.e("BLEProvisionLanding", "createSessionFailed!!" + LocalDateTime.now())
                        result.success(provisionStatus)
                    }

                    override fun wifiConfigSent() {
                        provisionStatus = "WIFI_CONFIG_SENT"
//                      sendProvisionResponse(method, provisionStatus!!)
                        println("wifiConfigSent!!" + LocalDateTime.now())
                        Log.e("BLEProvisionLanding", "wifiConfigSent!!" + LocalDateTime.now())
                    }

                    override fun wifiConfigFailed(e: java.lang.Exception) {
                        provisionStatus = "WIFI_CONFIG_FAILED"
//                      sendProvisionResponse(method, provisionStatus!!)
                        println("wifiConfigFailed!!" + LocalDateTime.now())
                        Log.e("BLEProvisionLanding", "wifiConfigFailed!!" + LocalDateTime.now())
                    }

                    override fun wifiConfigApplied() {
                        provisionStatus = "WIFI_CONFIG_APPLIED"
//                      sendProvisionResponse(method, provisionStatus!!)
                        println("wifiConfigApplied!!" + LocalDateTime.now())
                        Log.e("BLEProvisionLanding", "wifiConfigApplied!!" + LocalDateTime.now())
                    }

                    override fun wifiConfigApplyFailed(e: java.lang.Exception) {
                        provisionStatus = "WIFI_CONFIG_FAILED"
//                      sendProvisionResponse(method, provisionStatus!!)
                        println("wifiConfigApplyFailed!!" + LocalDateTime.now())
                        Log.e("BLEProvisionLanding", "wifiConfigApplyFailed!!" + LocalDateTime.now())
                    }

                    override fun provisioningFailedFromDevice(failureReason: ESPConstants.ProvisionFailureReason) {
                        provisionStatus = "DEVICE_PROVISIONING_FAILED"
//                      sendProvisionResponse(method, provisionStatus!!)
                        println("provisioningFailedFromDevice!!" + LocalDateTime.now())
                        Log.e("BLEProvisionLanding", "provisioningFailedFromDevice!!" + LocalDateTime.now())
                        Log.e("BLEProvisionLanding", failureReason.toString() + LocalDateTime.now())
                        result.success(provisionStatus)
                    }

                    override fun deviceProvisioningSuccess() {
                        provisionStatus = "DEVICE_PROVISIONING_SUCCESS"
//                      sendProvisionResponse(method, provisionStatus!!)
                        println("deviceProvisioningSuccess!!" + LocalDateTime.now())
                        Log.e("BLEProvisionLanding", "deviceProvisioningSuccess!!" + LocalDateTime.now())
                        result.success(provisionStatus)
                    }

                    override fun onProvisioningFailed(e: java.lang.Exception) {
                        provisionStatus = "DEVICE_PROVISIONING_FAILED"
//                      sendProvisionResponse(method, provisionStatus!!)
                        println("onProvisioningFailed!!" + LocalDateTime.now())
                        Log.e("BLEProvisionLanding", "onProvisioningFailed!!" + LocalDateTime.now())
                        result.success(provisionStatus)
                    }
                })

                println("Timer Starts" + LocalDateTime.now())
                Log.d("BLEProvisionLanding", "Timer Starts" + LocalDateTime.now())
            }

            method == "requestPermission" -> {
                Log.e("BLEProvisionLanding", "Request Perm")
                result.success(permissionManager.requestLocationPermission())
            }

            method == "permissionState" -> {
                result.success(permissionManager.bluetoothPermissionState)
            }

            method == "bluetoothState" -> {
                result.success(permissionManager.bluetoothState)
            }

            method.contains("bluetoothStateListener") -> {
                Log.e("BLEProvisionLanding", "bluetoothStateListener")

                permissionManager.bluetoothState(method, object : PermissionManager.BluetoothStateListener {
                    override fun onStateChanged(bluetoothState: PermissionManager.BluetoothState) {
                        sendBluetoothResponse(method, bluetoothState.state)
                        Log.e("BLEProvisionLanding", method)
                    }
                })
                result.success(true)
            }
        }
    }


    private fun sendBluetoothResponse(method: String, state: String) {
        val response: String = state
        Log.e("BLEProvisionLanding", method)
        Log.e("BLEProvisionLanding", state)
        channel.invokeMethod(method, response)
    }

    private fun sendDeviceResponse(method: String, devices: ArrayList<BleTooth>) {
        val response: HashMap<String, Any> = HashMap()
        val depdup: Set<BleTooth> = LinkedHashSet(devices)
        response["devices"] = depdup.map { b -> b.toObject() }
        Log.e("****BLEProvision*****", response["devices"].toString())
        channel.invokeMethod(method, response["devices"])
    }

    private fun sendProvisionResponse(method: String, status: String) {
//    val response: HashMap<String, Any> = HashMap()
        val response: String = status
        Log.e("BLEProvisionLanding", method)
        Log.e("BLEProvisionLanding", status)
        channel.invokeMethod(method, response)
    }

    override fun onDetachedFromEngine(@NonNull binding: FlutterPlugin.FlutterPluginBinding) {
        channel.setMethodCallHandler(null)
    }

    override fun onAttachedToActivity(binding: ActivityPluginBinding) {
        this.context = binding.activity
        initializeWithContext()
    }

    override fun onDetachedFromActivityForConfigChanges() {
        TODO("Not yet implemented")
    }

    override fun onReattachedToActivityForConfigChanges(binding: ActivityPluginBinding) {
        TODO("Not yet implemented")
    }

    override fun onDetachedFromActivity() {
        TODO("Not yet implemented")
    }
}


//enum class ProvisionState {
//  CREATE_SESSION_FAILED, WIFI_CONFIG_SENT, WIFI_CONFIG_FAILED, WIFI_CONFIG_APPLIED, PROVISION_FAILED_FROM_DEVICE, DEVICE_PROVISIONED, FAILED
//}






